/***************************************************************************************************
* Fox
* ---
* Based on vFox designed by VK6MHZ - www.ziotix.net/fox .
* Ver 1 18-10-18 - 6 beeps    - VK6TU - Michael Tutt - vk6tu.au@gmail.com
* Ver 2 08-11-18 - morse code - VK6TM - Kim Lyon     - kim.lyon@baremetal.systems
* 
****************************************************************************************************
* D4 PTT Pin, Low = TX, High = RX
* There is no PD Pin
* D0 TX Power Low = Low 0.5W, High = High 1W
* A0 Voltage Sensor also VCC
* D8 Serial RXD on Chip
* D7 Serial TXD On Chip
* D6 MIC Input
* D1/2 Are OLED
***************************************************************************************************/

/* ========================================== INCLUDES ========================================== */

#include <ESP8266HTTPClient.h>
#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <SoftwareSerial.h>
#include <FS.h>             // Include the SPIFFS library
#include "DRA818.h"
#include <Wire.h>           // Only needed for Arduino 1.6.5 and earlier
#include "SSD1306Wire.h"    // legacy include: `#include "SSD1306.h"`
#include "OLEDDisplay.h"

/* =========================================== ENUMS ============================================ */

enum
{
  CALL_SIGN_MAX_LEN = 16
};

/* ========================================== DEFINES =========================================== */

SSD1306Wire        display(0x3c, D1, D2);
#define PTT        D4
#define PD         D3
#define SOFTSERIAL 1
#define DRA_RXD    D7
#define DRA_TXD    D8
ESP8266WebServer   server(80);
SoftwareSerial     dra_serial(DRA_RXD, DRA_TXD);
DRA818             dra(&dra_serial, PTT); 

/* ========================================== VARIABLES ========================================= */

static int         counter        = 0 ;
static int         apready        = 0 ;
static const char* ssid           = "vFox" ;
static const char* password       = "winter11" ;
static long        previousMillis = 0 ; 
static int         aprunning ;
static String      power ;

/* ========================================== TYPEDEFS ========================================== */

typedef uint8_t  chr_t ;
typedef uint32_t period_t ;

/* ========================================= PROTOTYPES ========================================= */

static void TxIdent( chr_t* identMsg , period_t dotPeriod );

/* ========================================== FUNCTIONS ========================================= */

void setup( void ) 
{
  display.init();
  display.flipScreenVertically();
  display.setFont( ArialMT_Plain_10 );
  display.setTextAlignment( TEXT_ALIGN_LEFT );
  //                  x    y
  display.drawString( 0 , 10 , "Welcome to vFox"      );
  display.drawString( 0 , 20 , "Designed by VK6MHZ"   );
  display.drawString( 0 , 30 , "VHF/UHF Version 1W PA");
  display.drawString( 0 , 40 , "www.ziotix.net/fox"   );
  display.drawString( 0 , 50 , "Loading....."         );
  display.display();
  delay( 5000 );  

  pinMode( A0 , OUTPUT );
  pinMode( D0 , OUTPUT );
  pinMode( D4 , OUTPUT );
  dra_serial.begin( 9600 );
  Serial.begin( 9600 );
  Serial.println( "run" );
  SPIFFS.begin();

  delay( 1000 );
  String txfreqcontents ;
  
  // Get vars from SPIFFS
  File txfreq = SPIFFS.open( "FREQ" , "r" );
  
  if( txfreq ) 
  {
    String setfreq = txfreq.readString();
    Serial.println( setfreq );
    txfreq.close();
    float drafreq = setfreq.toFloat(); 

    dra.setFreq( drafreq );
    dra.setTXCTCSS( 0 );
    dra.setSquelch( 3 );
    dra.setRXCTCSS( 0 );
    dra.writeFreq();
    dra.setVolume( 4 );
    dra.setFilters( true , true , true );
 
  }

  File txpfile = SPIFFS.open( "TXP" , "r" );
  
  if( txpfile ) 
  {
    String txpower = ( txpfile.readString() );
    txpfile.close();

    if( txpower == "1" )
    {
      power = 1000 ;
      digitalWrite( D0 , HIGH );
    }
    else
    {
      power = 500 ;
      digitalWrite( D0 , LOW );
    }
  }

}

void loop( void ) 
{
  unsigned long cmillis = millis();

  server.handleClient();
  
  if( ( cmillis < 30000 ) && ( apready == 0 ) )
  {
    digitalWrite( PTT , HIGH );   // PTT is Active low, so this turns the PTT off.
    display.clear();
    //                  x    y
    display.drawString( 0 ,  0 , "Device in Config Mode"     );
    display.drawString( 0 ,  8 , "For Next 1 Minute"         );
    display.drawString( 0 , 16 , "Connect Wifi To Configure" );
    display.drawString( 0 , 24 , "SSID vFox"                 );
    display.drawString( 0 , 32 , "PASS winter11"             );
    display.drawString( 0 , 40 , "Web Server: 192.168.4.1"   );
    display.drawString( 0 , 50 , String( millis() )          );
    display.display();
    
    delay( 5000 );
    
    // not activated, Start WiFi AP
    WiFi.softAP( ssid , password );
    IPAddress myIP = WiFi.softAPIP();
    Serial.print( "AP IP address: " );
    Serial.println( myIP );
    server.on( "/" , configpage );
    server.on( "/take" , takeconfig );
    server.begin();
    Serial.println( "HTTP server started" ); 
    apready = 1 ;
    
  }
  
  if( ( cmillis > 60000 ) && ( apready == 1 ) )
  {
    transmit();
  }
  
}

void transmit( void )
{
  display.clear();
  File intFile = SPIFFS.open( "INT" , "r" );
  
  if( intFile ) 
  {
    String delaytime = intFile.readString();
    intFile.close();

    File mhzFile = SPIFFS.open( "FREQ" , "r" );
    
    if( mhzFile ) 
    {
      String mhz = mhzFile.readString();
      mhzFile.close();

      File wpmFile = SPIFFS.open( "WPM" , "r" );
      
      if( wpmFile ) 
      {
        String   wpmStr    = wpmFile.readString();
        wpmFile.close();
        uint32_t dotPeriod = 60000 / ( (uint32_t)wpmStr.toInt() * 50 ) ;
  
        File callsignFile = SPIFFS.open( "CALLSIGN" , "r" );
        
        if( callsignFile ) 
        {
          String   callsignStr = callsignFile.readString();
          callsignFile.close();

          uint8_t* callsign = (uint8_t*)callsignStr.c_str();

          File identFile = SPIFFS.open( "IDENT" , "r" );
        
          if( identFile ) 
          {
            String   identStr = identFile.readString();
            identFile.close();
  
            if( identStr.length() == 0 )
            {
              identStr = "CQ CQ CQ" ;
            }
            
            uint8_t* identMsg = (uint8_t*)identStr.c_str();
  
            Serial.println( "In Transmit Mode" );
            display.clear();
            //                   x    y
            display.drawString(  0 ,  0 ,              "Fox Running" );
            display.drawString(  0 ,  8 ,              "Frequency: " );
            display.drawString( 64 ,  8 ,              mhz           );
            display.drawString(  0 , 16 ,              "TX Power: "  );
            display.drawString( 64 , 16 ,              power         );
            display.drawString( 80 , 16 ,              "mW"          );
            display.drawString(  0 , 24 ,              "Timer: "     );
            display.drawString( 64 , 24 ,              delaytime     );
            display.drawString( 80 , 24 ,              "Sec"         );
            display.drawString(  0 , 32 ,              "Transmits: " );
            display.drawString(  0 , 40 ,              "Callsign: "  );
            display.drawString( 64 , 40 , (const char*)callsign      );
        
            String count ;
            count = String( counter );  
            counter++;
            display.drawString( 64 , 32 , count );
            display.display();
                      digitalWrite( D4 , LOW );   // PTT is Active low, so this turns the PTT on.
            delay( 250 );
            
            TxIdent( identMsg , dotPeriod );
            
            digitalWrite( D4 , HIGH ); // PTT off.
            
            int dtime = 1000 * delaytime.toInt();
            delay( dtime );

          }
        }
      }
    }
  }
  
}

void configpage( void ) 
{
  server.send( 200 , "text/html",
"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">"
"<HTML>"
  "<HEAD>"
    "<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=iso-8859-1\">"
    "<meta name=\"generator\" content=\"Bluefish 2.2.7\" >"
    "<TITLE>vFox</TITLE>"
    "<style type=\"text/css\">"
      ".page { background-color: #FDFDDA; margin-top: 20; margin-left: 20; }"
      ".parameter_table { width: 800; }"
      ".parameter_table_row { text-align: left; vertical-align: top; }"
      ".parameter_table_text_cell { text-align: left; vertical-align: top; width: 280; }"
      ".parameter_table_setting_cell { text-align: left; vertical-align: top; width: 100; }"
      ".parameter_table_setting { text-align: left; vertical-align: top; width: 100; }"
      ".parameter_table_ident_cell { text-align: left; vertical-align: top; width: 520; height: 80; }"
      ".parameter_table_ident { text-align: left; vertical-align: top; width: 520; height: 80; }"
      ".parameter_table_submit_cell { text-align: center; vertical-align: middle; width: 520; }"
      ".parameter_table_submit { text-align: center; vertical-align: middle; width: 520; }"
      ".help_table { width: 800; }"
      ".help_table_row { text-align: left; vertical-align: top; height: 25; }"
      ".help_table_instruction_cell { text-align: left; vertical-align: top; width: 100; }"
      ".help_table_description_cell { text-align: middle; vertical-align: top; }"
    "</style>"
  "</HEAD>"
  "<BODY class=\"page\">"
    "<h1>Welcome to the Fox</h1>"
    "Please enter Configuration Details below, Once Complete click on Set.<br><br>"
    "<form action=/take method=post>"
      "<table class=\"parameter_table\">"
        "<tr class=\"parameter_table_row\"><td class=\"parameter_table_text_cell\">Frequency in Mhz, Example: 433.000:</td><td class=\"parameter_table_setting_cell\"><input type=text name=freq class=\"parameter_table_setting\"></td><td>&nbsp;</td></tr>"
        "<tr class=\"parameter_table_row\"><td class=\"parameter_table_text_cell\">Power 1 for High, 0.5 for low</td><td class=\"parameter_table_setting_cell\"><input type=text name=txpower class=\"parameter_table_setting\"></td><td>&nbsp;</td></tr>"
        "<tr class=\"parameter_table_row\"><td class=\"parameter_table_text_cell\">Interval in Seconds</td><td class=\"parameter_table_setting_cell\"><input type=text name=interval class=\"parameter_table_setting\"></td><td>&nbsp;</td></tr>"
        "<tr class=\"parameter_table_row\"><td class=\"parameter_table_text_cell\">WPM</td><td class=\"parameter_table_setting_cell\"><input type=text name=wpm class=\"parameter_table_setting\"></td><td>&nbsp;</td></tr>"
        "<tr class=\"parameter_table_row\"><td class=\"parameter_table_text_cell\">Callsign</td><td class=\"parameter_table_setting_cell\"><input type=text name=callsign class=\"parameter_table_setting\"></td><td>&nbsp;</td></tr>"
        "<tr class=\"parameter_table_row\"><td class=\"parameter_table_text_cell\">Ident<br>eg. callsign<br>followed by beeps</td>"
          "<td colspan=\"2\" class=\"parameter_table_ident_cell\"><textarea type=text name=ident class=\"parameter_table_ident\"></textarea></td>"
        "</tr>"
        "<tr class=\"parameter_table_row\"><td class=\"parameter_table_text_cell\">&nbsp;</td><td colspan=\"2\" class=\"parameter_table_submit_cell\"><input type=submit value=Set class=\"parameter_table_submit\"></td></tr>"
      "</table>"
    "</form>"
    "<u>In the Ident</u><br>"
    "<table class=\"help_table\">"
      "<tr class=\"help_table_row\"><td class=\"help_table_instruction_cell\">[2000]</td><td class=\"help_table_description_cell\">will wait for 2000ms and store this value.</td></tr>"
      "<tr class=\"help_table_row\"><td class=\"help_table_instruction_cell\">#[2000]</td><td class=\"help_table_description_cell\">will only load the wait period of 2000ms and store this value.</td></tr>"
      "<tr class=\"help_table_row\"><td class=\"help_table_instruction_cell\">$</td><td class=\"help_table_description_cell\">will wait for stored value of wait period.</td></tr>"
      "<tr class=\"help_table_row\"><td class=\"help_table_instruction_cell\">&lt;1000&gt;</td><td class=\"help_table_description_cell\">will transmit a beep for 1000ms and store this value.</td></tr>"
      "<tr class=\"help_table_row\"><td width=100>#&lt;1000&gt;</td><td class=\"help_table_description_cell\">will only load the beep period of 1000ms and store this value.</td></tr>"
      "<tr class=\"help_table_row\"><td class=\"help_table_instruction_cell\">*</td><td class=\"help_table_description_cell\">will transmit a beep for stored value of beep period.<br>"
                         "eg. [200]<300>$*$* will transmit a sequence of 3 x 200ms wait followed by 300ms beep<br>"
                         "&nbsp;&nbsp;&nbsp;&nbsp;#[200]#<300>$*$*$* will transmit a sequence of 3 x 200ms wait followed by 300ms beep</td></tr>"
      "<tr class=\"help_table_row\"><td class=\"help_table_instruction_cell\">(5)</td><td class=\"help_table_description_cell\">will set the number of repeats to 5 and store this value.</td></tr>"
      "<tr class=\"help_table_row\"><td class=\"help_table_instruction_cell\">{ }</td><td class=\"help_table_description_cell\">will repeat the instructions inbetween for stored number of repeats.<br>eg. (5){[200]<300>} will transmit a repeated sequence of 5 x 200ms wait followed by 300ms beep.</td></tr>"
    "</table>"
  "</BODY>"
"</HTML>"
    );
}

void takeconfig( void ) 
{
  server.send( 200 , "text/html" , "<h1>Please Wait while vFox saves your settings...</h1>" );

  String frequency = server.arg( "freq"     );
  String txp       = server.arg( "txpower"  );
  String inter     = server.arg( "interval" );
  String wpm       = server.arg( "wpm"      );
  String callsign  = server.arg( "callsign" );
  String ident     = server.arg( "ident"    );
  
  File TXPFile = SPIFFS.open( "TXP" , "w" );
  TXPFile.print( txp );
  TXPFile.close();
  
  File FREQFile = SPIFFS.open( "FREQ" , "w" );
  FREQFile.print( frequency );
  FREQFile.close();
  
  File INTFile = SPIFFS.open( "INT" , "w" );
  INTFile.print( inter );
  INTFile.close();
  
  File WPMFile = SPIFFS.open( "WPM" , "w" );
  WPMFile.print( wpm );
  WPMFile.close();
  
  File CALLSIGNFile = SPIFFS.open( "CALLSIGN" , "w" );
  CALLSIGNFile.print( callsign );
  CALLSIGNFile.close();
  
  File IDENTFile = SPIFFS.open( "IDENT" , "w" );
  IDENTFile.print( ident );
  IDENTFile.close();

}

/***************************************************************************************************
* Code below and associated code is by :-
* ---------------------------------------
* Kim Lyon
* BareMetal.Systems
* (c) copyright is by Bare Metal Systems and all rights are reserved.
* The code is supplied as freeware and as such may be freely copied, modified and used.
***************************************************************************************************/

/* =========================================== ENUMS ============================================ */

// base time lengths
enum
{
    CW_PERIOD_DOT            = 1                 ,
    CW_PERIOD_DASH           = 3 * CW_PERIOD_DOT ,
    CW_PERIOD_DOT_DASH_SPACE =     CW_PERIOD_DOT ,
    CW_PERIOD_LETTER_SPACE   = 3 * CW_PERIOD_DOT ,
    CW_PERIOD_WORD_SPACE     = 7 * CW_PERIOD_DOT 
};

/***************************************************************************************************
* Morse Code Character Table
***************************************************************************************************/

/* dot \ dash \ command */
enum
{
    CW_NULL        = 0x00 ,
    CW_DOT         = 0x01 ,
    CW_DASH        = 0x02 , 
    CW_CMD         = 0x03 , 
    CW_MASK        = 0x03 ,
    CW_SHIFT       = 2    ,
    CW_CMD_MASK    = 0x0F ,
    CW_NUM_OF_CMDS = CW_CMD_MASK + 1 
};

// Commands
// Corresponds to Cmds table.
enum
{
  CW_CMD_NOP        ,
  CW_CMD_LOAD_ONLY  , // #
  CW_CMD_LOAD_VALUE , // < or [ or (
  CW_CMD_WAIT_N     , // ]
  CW_CMD_WAIT       , // $
  CW_CMD_BEEP_N     , // >
  CW_CMD_BEEP       , // *
  CW_CMD_REPEAT_NUM , // )
  CW_CMD_NEST_START , // {
  CW_CMD_NEST_END     // }
};

/* individual dots \ dashes */
enum
{
    CW_0_SHIFT =  0                    ,
    CW_0_NULL  = CW_NULL << CW_0_SHIFT ,
    CW_0_DOT   = CW_DOT  << CW_0_SHIFT ,
    CW_0_DASH  = CW_DASH << CW_0_SHIFT ,

    CW_1_SHIFT = CW_0_SHIFT + CW_SHIFT ,
    CW_1_NULL  = CW_NULL << CW_1_SHIFT ,
    CW_1_DOT   = CW_DOT  << CW_1_SHIFT ,
    CW_1_DASH  = CW_DASH << CW_1_SHIFT , 
    
    CW_2_SHIFT = CW_1_SHIFT + CW_SHIFT ,
    CW_2_NULL  = CW_NULL << CW_2_SHIFT ,
    CW_2_DOT   = CW_DOT  << CW_2_SHIFT ,
    CW_2_DASH  = CW_DASH << CW_2_SHIFT , 
    
    CW_3_SHIFT = CW_2_SHIFT + CW_SHIFT ,
    CW_3_NULL  = CW_NULL << CW_3_SHIFT ,
    CW_3_DOT   = CW_DOT  << CW_3_SHIFT ,
    CW_3_DASH  = CW_DASH << CW_3_SHIFT , 
    
    CW_4_SHIFT = CW_3_SHIFT + CW_SHIFT ,
    CW_4_NULL  = CW_NULL << CW_4_SHIFT ,
    CW_4_DOT   = CW_DOT  << CW_4_SHIFT ,
    CW_4_DASH  = CW_DASH << CW_4_SHIFT , 
    
    CW_5_SHIFT = CW_4_SHIFT + CW_SHIFT ,
    CW_5_NULL  = CW_NULL << CW_5_SHIFT ,
    CW_5_DOT   = CW_DOT  << CW_5_SHIFT ,
    CW_5_DASH  = CW_DASH << CW_5_SHIFT , 
    
    CW_6_SHIFT = CW_5_SHIFT + CW_SHIFT ,
    CW_6_NULL  = CW_NULL << CW_6_SHIFT ,
    CW_6_DOT   = CW_DOT  << CW_6_SHIFT ,
    CW_6_DASH  = CW_DASH << CW_6_SHIFT , 
    
    CW_7_SHIFT = CW_6_SHIFT + CW_SHIFT ,
    CW_7_NULL  = CW_NULL << CW_7_SHIFT ,
    CW_7_DOT   = CW_DOT  << CW_7_SHIFT ,
    CW_7_DASH  = CW_DASH << CW_7_SHIFT 
    
};

enum
{
    CW_MAX_NUM_DOTS_DASHES = 8 ,
    CW_TOP_SHIFT           = CW_7_SHIFT 
};

/* =========================================== MACROS =========================================== */

/* dot \ dash */
#define CW( h , g , f , e , d , c , b , a )   CW_7_##h | CW_6_##g | CW_5_##f | CW_4_##e | CW_3_##d | CW_2_##c | CW_1_##b | CW_0_##a  
// command
#define CMD( cmd )                            ( CW_CMD << CW_TOP_SHIFT ) | CW_CMD_##cmd

/* ========================================== TYPEDEFS ========================================== */

typedef uint16_t cw_t ;

/* ========================================== CONSTANTS ========================================= */

// Character to Morse Code translation table
// Uses 8 x  2 bit fields.
static cw_t CWCode[ 128 ] =
{
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x00
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x01
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x02
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x03
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x04
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x05
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x06
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x07
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x08
    CMD( NOP                                                  ) , // 0x09 TAB
    CMD( NOP                                                  ) , // 0x0A LF
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x0B
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x0C
    CMD( NOP                                                  ) , // 0x0D CR
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x0E
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x0F
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x10
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x11
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x12
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x13
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x14
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x15
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x16
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x17
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x18
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x19
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x1A
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x1B ESC
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x1C
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x1D
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x1E
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x1F
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x20
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x21 !
    CW( DOT  , DASH , DOT  , DOT  , DASH , DOT  , NULL , NULL ) , // 0x22 " . _ . . _ .
    CMD( LOAD_ONLY                                            ) , // 0x23 #
    CMD( WAIT                                                 ) , // 0x24 $
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x25 %
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x26 &
    CW( DOT  , DASH , DASH , DASH , DASH , DOT  , NULL , NULL ) , // 0x27 ' . _ _ _ _ .
    CMD( LOAD_VALUE                                           ) , // 0x28 ( _ . _ _ .
    CMD( REPEAT_NUM                                           ) , // 0x29 ) _ . _ _ . _
    CMD( BEEP                                                 ) , // 0x2A * _ . . _
    CW( DOT  , DASH , DOT  , DASH , DOT  , NULL , NULL , NULL ) , // 0x2B + . _ . _ .
    CW( DASH , DASH , DOT  , DOT  , DASH , DASH , NULL , NULL ) , // 0x2C , _ _ . . _ _
    CW( DASH , DOT  , DOT  , DOT  , DOT  , DASH , NULL , NULL ) , // 0x2D - _ . . . . _
    CW( DOT  , DOT  , DASH , DOT  , DASH , DOT  , DASH , NULL ) , // 0x2E . . _ . _ . _
    CW( DASH , DOT  , DOT  , DASH , DOT  , NULL , NULL , NULL ) , // 0x2F / _ . . _ .
    CW( DASH , DASH , DASH , DASH , DASH , NULL , NULL , NULL ) , // 0x30 0 _ _ _ _ _
    CW( DOT  , DASH , DASH , DASH , DASH , NULL , NULL , NULL ) , // 0x31 1 . _ _ _ _
    CW( DOT  , DOT  , DASH , DASH , DASH , NULL , NULL , NULL ) , // 0x32 2 . . _ _ _
    CW( DOT  , DOT  , DOT  , DASH , DASH , NULL , NULL , NULL ) , // 0x33 3 . . . _ _
    CW( DOT  , DOT  , DOT  , DOT  , DASH , NULL , NULL , NULL ) , // 0x34 4 . . . . _
    CW( DOT  , DOT  , DOT  , DOT  , DOT  , NULL , NULL , NULL ) , // 0x35 5 . . . . .
    CW( DASH , DOT  , DOT  , DOT  , DOT  , NULL , NULL , NULL ) , // 0x36 6 _ . . . .
    CW( DASH , DASH , DOT  , DOT  , DOT  , NULL , NULL , NULL ) , // 0x37 7 _ _ . . .
    CW( DASH , DASH , DASH , DOT  , DOT  , NULL , NULL , NULL ) , // 0x38 8 _ _ _ . .
    CW( DASH , DASH , DASH , DASH , DOT  , NULL , NULL , NULL ) , // 0x39 9 _ _ _ _ .
    CW( DASH , DASH , DASH , DOT  , DOT  , DOT  , NULL , NULL ) , // 0x3A : _ _ _ . . .
    CW( DASH , DOT  , DASH , DOT  , DASH , DOT  , NULL , NULL ) , // 0x3B ; _ . _ . _ .
    CMD( LOAD_VALUE                                           ) , // 0x3C <
    CW( DASH , DOT  , DOT  , DOT  , DASH , NULL , NULL , NULL ) , // 0x3D = _ . . . _
    CMD( BEEP_N                                               ) , // 0x3E >
    CW( DOT  , DOT  , DASH , DASH , DOT  , DOT  , NULL , NULL ) , // 0x3F ? . . _ _ . .
    CW( DOT  , DASH , DASH , DOT  , DASH , DOT  , NULL , NULL ) , // 0x40 @ . _ _ . _ .
    CW( DOT  , DASH , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x41 A . _
    CW( DASH , DOT  , DOT  , DOT  , NULL , NULL , NULL , NULL ) , // 0x42 B _ . . .
    CW( DASH , DOT  , DASH , DOT  , NULL , NULL , NULL , NULL ) , // 0x43 C _ . _ .
    CW( DASH , DOT  , DOT  , NULL , NULL , NULL , NULL , NULL ) , // 0x44 D _ . .
    CW( DOT  , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x45 E .
    CW( DOT  , DOT  , DASH , DOT  , NULL , NULL , NULL , NULL ) , // 0x46 F . . _ .
    CW( DASH , DASH , DOT  , NULL , NULL , NULL , NULL , NULL ) , // 0x47 G _ _ .
    CW( DOT  , DOT  , DOT  , DOT  , NULL , NULL , NULL , NULL ) , // 0x48 H . . . .
    CW( DOT  , DOT  , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x49 I . .
    CW( DOT  , DASH , DASH , DASH , NULL , NULL , NULL , NULL ) , // 0x4A J . _ _ _
    CW( DASH , DOT  , DASH , NULL , NULL , NULL , NULL , NULL ) , // 0x4B K _ . _
    CW( DOT  , DASH , DOT  , DOT  , NULL , NULL , NULL , NULL ) , // 0x4C L . _ . .
    CW( DASH , DASH , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x4D M _ _
    CW( DASH , DOT  , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x4E N _ .
    CW( DASH , DASH , DASH , NULL , NULL , NULL , NULL , NULL ) , // 0x4F O _ _ _
    CW( DOT  , DASH , DASH , DOT  , NULL , NULL , NULL , NULL ) , // 0x50 P . _ _ .
    CW( DASH , DASH , DOT  , DASH , NULL , NULL , NULL , NULL ) , // 0x51 Q _ _ . _
    CW( DOT  , DASH , DOT  , NULL , NULL , NULL , NULL , NULL ) , // 0x52 R . _ .
    CW( DOT  , DOT  , DOT  , NULL , NULL , NULL , NULL , NULL ) , // 0x53 S . . .
    CW( DASH , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x54 T _
    CW( DOT  , DOT  , DASH , NULL , NULL , NULL , NULL , NULL ) , // 0x55 U . . _
    CW( DOT  , DOT  , DOT  , DASH , NULL , NULL , NULL , NULL ) , // 0x56 V . . . _
    CW( DOT  , DASH , DASH , NULL , NULL , NULL , NULL , NULL ) , // 0x57 W . _ _
    CW( DASH , DOT  , DOT  , DASH , NULL , NULL , NULL , NULL ) , // 0x58 X _ . . _
    CW( DASH , DOT  , DASH , DASH , NULL , NULL , NULL , NULL ) , // 0x59 Y _ . _ _
    CW( DASH , DASH , DOT  , DOT  , NULL , NULL , NULL , NULL ) , // 0x5A Z _ _ . .
    CMD( LOAD_VALUE                                           ) , // 0x5B [
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , /* 0x5C \ */
    CMD( WAIT_N                                               ) , // 0x5D ]
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x5E ^
    CW( DASH , DOT  , DOT  , DASH , DASH , DOT  , DASH , NULL ) , // 0x5F _ . . _ _ . _
    CW( DOT  , DASH , DASH , DASH , DASH , DOT  , NULL , NULL ) , // 0x60 `
    CW( DOT  , DASH , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x61 a . _
    CW( DASH , DOT  , DOT  , DOT  , NULL , NULL , NULL , NULL ) , // 0x62 b _ . . .
    CW( DASH , DOT  , DASH , DOT  , NULL , NULL , NULL , NULL ) , // 0x63 c _ . _ .
    CW( DASH , DOT  , DOT  , NULL , NULL , NULL , NULL , NULL ) , // 0x64 d _ . .
    CW( DOT  , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x65 e .
    CW( DOT  , DOT  , DASH , DOT  , NULL , NULL , NULL , NULL ) , // 0x66 f . . _ .
    CW( DASH , DASH , DOT  , NULL , NULL , NULL , NULL , NULL ) , // 0x67 g _ _ .
    CW( DOT  , DOT  , DOT  , DOT  , NULL , NULL , NULL , NULL ) , // 0x68 h . . . .
    CW( DOT  , DOT  , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x69 i . .
    CW( DOT  , DASH , DASH , DASH , NULL , NULL , NULL , NULL ) , // 0x6A j . _ _ _
    CW( DASH , DOT  , DASH , NULL , NULL , NULL , NULL , NULL ) , // 0x6B k _ . _
    CW( DOT  , DASH , DOT  , DOT  , NULL , NULL , NULL , NULL ) , // 0x6C l . _ . .
    CW( DASH , DASH , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x6D m _ _
    CW( DASH , DOT  , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x6E n _ .
    CW( DASH , DASH , DASH , NULL , NULL , NULL , NULL , NULL ) , // 0x6F o _ _ _
    CW( DOT  , DASH , DASH , DOT  , NULL , NULL , NULL , NULL ) , // 0x70 p . _ _ .
    CW( DASH , DASH , DOT  , DASH , NULL , NULL , NULL , NULL ) , // 0x71 q _ _ . _
    CW( DOT  , DASH , DOT  , NULL , NULL , NULL , NULL , NULL ) , // 0x72 r . _ .
    CW( DOT  , DOT  , DOT  , NULL , NULL , NULL , NULL , NULL ) , // 0x73 s . . .
    CW( DASH , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x74 t _
    CW( DOT  , DOT  , DASH , NULL , NULL , NULL , NULL , NULL ) , // 0x75 u . . _
    CW( DOT  , DOT  , DOT  , DASH , NULL , NULL , NULL , NULL ) , // 0x76 v . . . _
    CW( DOT  , DASH , DASH , NULL , NULL , NULL , NULL , NULL ) , // 0x77 w . _ _
    CW( DASH , DOT  , DOT  , DASH , NULL , NULL , NULL , NULL ) , // 0x78 x _ . . _
    CW( DASH , DOT  , DASH , DASH , NULL , NULL , NULL , NULL ) , // 0x79 y _ . _ _
    CW( DASH , DASH , DOT  , DOT  , NULL , NULL , NULL , NULL ) , // 0x7A z _ _ . .
    CMD( NEST_START                                           ) , // 0x7B {
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x7C |
    CMD( NEST_END                                             ) , // 0x7D }
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL ) , // 0x7E ~
    CW( NULL , NULL , NULL , NULL , NULL , NULL , NULL , NULL )   // 0x7F DEL
};

/***************************************************************************************************
* Parameters
***************************************************************************************************/

enum
{
  CW_PARA_FREQ                =  500 ,
  CW_PARA_DOT_PERIOD_DEFAULT  =  100 ,
  CW_PARA_MAX_MSG_SIZE        =  255 ,

  CW_PARA_REPEAT_NUM_DEFAULT  =    5 ,
  CW_PARA_WAIT_PERIOD_DEFAULT = 1000 ,
  CW_PARA_BEEP_PERIOD_DEFAULT = 5000 
  
};

/* ========================================== TYPEDEFS ========================================== */

typedef uint8_t  bool_t ;

typedef uint32_t repeat_t ;
typedef uint16_t index_t ;
typedef bool_t   loadonly_t ;
typedef uint32_t value_t ;

/* ========================================== STRUCTS =========================================== */

// ident structure - object
typedef struct ident_tag
{
  chr_t*     msg ;          // pointer to the ident message being transmitted
  index_t    index ;        // pointer to the message index - current position - advanced
  loadonly_t loadOnly ;     // indicates whether to execute the command or to only load the associated value
  value_t    value ;        // the value - returned or used
  period_t   dotPeriod ;    // the period of dots in ms - basic timing - derived from WPM
  period_t   waitPeriod ;   // the wait period in ms - stored value
  period_t   beepPeriod ;   // the beep period in ms - stored value
  repeat_t   repeatNum ;    // number of repeates of nested instructions
  repeat_t   currRepeat ;   // the current repeat number
  index_t    nestStart ;    // the start of the nest
}ident_s;

/* ========================================== TYPEDEFS ========================================== */

typedef void (*cmd_f)( ident_s* ident );

/* ========================================= PROTOTYPES ========================================= */

static void TxChr( chr_t chr , period_t dotPeriod );
static void TxTone( period_t period );
static void Delay( period_t period );
// commands
static void Cmd_NOP( ident_s* ident );
static void Cmd_LoadOnly( ident_s* ident );
static void Cmd_LoadValue( ident_s* ident );
static void Cmd_WaitN( ident_s* ident );
static void Cmd_Wait( ident_s* ident );
static void Cmd_BeepN( ident_s* ident );
static void Cmd_Beep( ident_s* ident );
static void Cmd_RepeatNum( ident_s* ident );
static void Cmd_NestStart( ident_s* ident );
static void Cmd_NestEnd( ident_s* ident );

/* ========================================== CONSTANTS ========================================= */

// commands
// Corresponds to CW_CMD_... commands.
static const cmd_f Cmds[ CW_NUM_OF_CMDS ] =
{
  Cmd_NOP       , // 0x00
  Cmd_LoadOnly  , // 0x01
  Cmd_LoadValue , // 0x02
  Cmd_WaitN     , // 0x03
  Cmd_Wait      , // 0x04
  Cmd_BeepN     , // 0x05
  Cmd_Beep      , // 0x06
  Cmd_RepeatNum , // 0x07
  Cmd_NestStart , // 0x08
  Cmd_NestEnd   , // 0x09
  Cmd_NOP       , // 0x0A
  Cmd_NOP       , // 0x0B
  Cmd_NOP       , // 0x0C
  Cmd_NOP       , // 0x0D
  Cmd_NOP       , // 0x0E
  Cmd_NOP         // 0x0F
};

/* ========================================== FUNCTIONS ========================================= */

/***************************************************************************************************
* Function    : TxIdent
* 
* Description : Transmit the message in morse. Execute any specified commands.
* 
* Parameters  : chr_t*    identMsg    Pointer to the ident message to transmit.
*               period_t  dotPeriod   The period of dots in ms - basic timing.
*              
* Return      : none
*
***************************************************************************************************/

static void TxIdent( chr_t* identMsg , period_t dotPeriod )
{
  ident_s ident ;
  chr_t   chr ;
  cw_t    bits ;
  cw_t    cmd ;

  ident.msg        = identMsg ;
  ident.index      = 0 ;
  ident.loadOnly   = false ;
  ident.value      = 0 ;
  ident.dotPeriod  = dotPeriod;
  ident.repeatNum  = CW_PARA_REPEAT_NUM_DEFAULT ;
  ident.currRepeat = 0 ;
  ident.waitPeriod = CW_PARA_WAIT_PERIOD_DEFAULT ;
  ident.beepPeriod = CW_PARA_BEEP_PERIOD_DEFAULT ;
  ident.nestStart  = 0 ;

  for( ident.index = 0 ; ident.msg[ ident.index ] && ( ident.index < CW_PARA_MAX_MSG_SIZE ) ; ident.index++ )
  {
    chr  = ident.msg[ ident.index ] ;
    bits = CWCode[ chr ] ;
    
    if( ( ( bits >> CW_TOP_SHIFT ) & CW_MASK ) == CW_CMD )
    {
      ident.index++ ;
      cmd = bits & CW_CMD_MASK ;
      Cmds[ cmd ]( &ident );
      ident.index-- ;
    }
    else
    {
        if( chr == ' ' )
        {
          Delay( ( CW_PERIOD_WORD_SPACE - CW_PERIOD_DOT_DASH_SPACE ) * ident.dotPeriod );
        }
        else
        {
          TxChr( chr , ident.dotPeriod );
          Delay( ( CW_PERIOD_LETTER_SPACE - CW_PERIOD_DOT_DASH_SPACE ) * ident.dotPeriod );
        }
       
    }

  }

}

/***************************************************************************************************
* Commands
***************************************************************************************************/

/***************************************************************************************************
* Description : Execute the specified commands.
* 
* Parameters  : ident_s*    ident   Pointer to the ident object.
*              
* Return      : none
*
***************************************************************************************************/

// no operation
static void Cmd_NOP( ident_s* ident )
{
}

// load value only
static void Cmd_LoadOnly( ident_s* ident )
{
    ident->loadOnly = true ;
}

// load value
static void Cmd_LoadValue( ident_s* ident )
{
  // load decimal numeric value
  for( ident->value = 0 ; 
       ident->msg[ ident->index ] && ( ident->index < CW_PARA_MAX_MSG_SIZE ) && 
       ( ident->msg[ ident->index ] >= '0' ) && ( ident->msg[ ident->index ] <= '9' ) ; 
       ident->index++                                                                   )
  {
    ident->value *= 10 ;
    ident->value += (uint32_t)( ident->msg[ ident->index ] - '0' ) ;
  }
  
}

// wait for Ns
static void Cmd_WaitN( ident_s* ident )
{
  ident->waitPeriod = ident->value ;
  Delay( ident->waitPeriod );
}

// wait
static void Cmd_Wait( ident_s* ident )
{
  if( !ident->loadOnly )
  {
    Delay( ident->waitPeriod );
  }
  ident->loadOnly = false ;
}

// beep for Ns - transmit a long tone
static void Cmd_BeepN( ident_s* ident )
{
  ident->beepPeriod = ident->value ;
  TxTone( ident->beepPeriod );
}

// beep 
static void Cmd_Beep( ident_s* ident )
{
  if( !ident->loadOnly )
  {
    TxTone( ident->beepPeriod );
  }
  ident->loadOnly = false ;
}

// repeat num
static void Cmd_RepeatNum( ident_s* ident )
{
  ident->repeatNum = ident->value ;
}

// nest start
static void Cmd_NestStart( ident_s* ident )
{
  ident->currRepeat = 0 ;
  ident->nestStart  = ident->index ;
}

// nest end
static void Cmd_NestEnd( ident_s* ident )
{
  ident->currRepeat++ ;
  
  if( ident->currRepeat < ident->repeatNum )
  {
    ident->index = ident->nestStart ;
  }
  
}

/***************************************************************************************************
* Component Functions
***************************************************************************************************/

/***************************************************************************************************
* Function    : TxChr
* 
* Description : Transmit the character in morse.
* 
* Parameters  : chr_t     chr         The character to be transmitted in morse.
*               period_t  dotPeriod   The period of dots in ms - basic timing.
*              
* Return      : none
*
***************************************************************************************************/

static void TxChr( chr_t chr , period_t dotPeriod )
{ 
  period_t period ;
  index_t  index ;
  index_t  shift ;
  cw_t     bits ;

  for( index = 0                                     ,
       shift = CW_TOP_SHIFT                          ,
       bits  = ( CWCode[ chr ] >> shift ) & CW_MASK    ;
       ( index < CW_MAX_NUM_DOTS_DASHES ) && 
       ( bits & ( CW_DOT | CW_DASH )    )              ; 
       index++                                       , 
       shift -= CW_SHIFT                             , 
       bits   = ( CWCode[ chr ] >> shift ) & CW_MASK   )
  {
    if( bits & CW_DOT )
    {
      TxTone( CW_PERIOD_DOT * dotPeriod );
    }
    else
    {
      if( bits & CW_DASH )
      {
        TxTone( CW_PERIOD_DASH * dotPeriod );
      }
    }

    Delay( CW_PERIOD_DOT_DASH_SPACE * dotPeriod );
    
  }

}

/***************************************************************************************************
* Function    : TxTone
* 
* Description : Transmit the tone.
* 
* Parameters  : period_t  period  The length of the tone in ms.
*              
* Return      : none
*
***************************************************************************************************/

static void TxTone( period_t period )
{
  tone( D6 , CW_PARA_FREQ ); 
  delay( period );
  noTone( D6 ); 
}

/***************************************************************************************************
* Function    : Delay
* 
* Description : Wait for the specified period.
* 
* Parameters  : period_t  period  The length of the wait in ms.
*              
* Return      : none
*
***************************************************************************************************/

static void Delay( period_t period )
{
  delay( period );
}
